# Generated by Django 2.0.1 on 2018-02-27 21:08

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('auth', '0009_alter_user_last_name_max_length'),
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='RolePermission',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('access', models.BooleanField(choices=[(True, 'Allow'), (False, 'Deny')], default=True)),
            ],
        ),
        migrations.CreateModel(
            name='UserRole',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('role_class', models.CharField(max_length=256)),
                ('object_id', models.PositiveIntegerField(null=True)),
                ('content_type', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
                ('permissions', models.ManyToManyField(related_name='roles', through='improved_permissions.RolePermission', to='auth.Permission', verbose_name='Permissões')),
            ],
            options={
                'verbose_name_plural': 'Role Instances',
                'verbose_name': 'Role Instance',
            },
        ),
    ]



"""assignments functions"""
from django.contrib.contenttypes.models import ContentType

from improved_permissions.exceptions import (InvalidPermissionAssignment,
                                             InvalidRoleAssignment)
from improved_permissions.getters import get_roles, get_users
from improved_permissions.models import RolePermission, UserRole
from improved_permissions.roles import ALL_MODELS
from improved_permissions.utils import (check_my_model, delete_from_cache,
                                        get_roleclass, is_unique_together,
                                        string_to_permission)


def assign_role(user, role_class, obj=None):
    """
    Proxy method to be used for one
    User instance.
    """
    assign_roles([user], role_class, obj)


def assign_roles(users_list, role_class, obj=None):
    """
    Create a RolePermission object in the database
    referencing the followling role_class to the
    user.
    """
    users_set = set(users_list)
    role = get_roleclass(role_class)
    name = role.get_verbose_name()

    # Check if object belongs
    # to the role class.
    check_my_model(role, obj)

    # If no object is provided but the role needs specific models.
    if not obj and role.models != ALL_MODELS:
        raise InvalidRoleAssignment(
            'The role "%s" must be assigned with a object.' % name
        )

    # If a object is provided but the role does not needs a object.
    if obj and role.models == ALL_MODELS:
        raise InvalidRoleAssignment(
            'The role "%s" must not be assigned with a object.' % name
        )

    # Check if the model accepts multiple roles
    # attached using the same User instance.
    if obj and is_unique_together(obj):
        for user in users_set:
            has_user = get_roles(user=user, obj=obj)
            if has_user:
                raise InvalidRoleAssignment(
                    'The user "%s" already has a role attached '
                    'to the object "%s".' % (user, obj)
                )

    if role.unique is True:
        # If the role is marked as unique but multiple users are provided.
        if len(users_list) > 1:
            raise InvalidRoleAssignment(
                'Multiple users were provided using "%s", '
                'but it is marked as unique.' % name
            )

        # If the role is marked as unique but already has an user attached.
        has_user = get_users(role_class=role, obj=obj)
        if has_user:
            raise InvalidRoleAssignment(
                'The object "%s" already has a "%s" attached '
                'and it is marked as unique.' % (obj, name)
            )

    for user in users_set:
        ur_instance = UserRole(role_class=role.get_class_name(), user=user)
        if obj:
            ur_instance.obj = obj
        ur_instance.save()

        # Cleaning the cache system.
        delete_from_cache(user, obj)


def assign_permission(user, role_class, permission, access, obj=None):
    """
    Assign a specific permission value
    to a given UserRole instance.
    The values used in this method overrides
    any configuration of "allow/deny" or
    "inherit_allow/inherit_deny".
    """
    role = get_roleclass(role_class)
    perm = string_to_permission(permission)
    query = UserRole.objects.filter(user=user, role_class=role.get_class_name())
    if obj:
        ct_obj = ContentType.objects.get_for_model(obj)
        query = query.filter(content_type=ct_obj.id, object_id=obj.id)

    if not query:
        raise InvalidPermissionAssignment('No Role instance was affected.')

    for role_obj in query:
        perm_obj, created = RolePermission.objects.get_or_create(  # pylint: disable=W0612
            role=role_obj,
            permission=perm
        )
        perm_obj.access = bool(access)
        perm_obj.save()

        # Cleaning the cache system.
        delete_from_cache(user, role_obj.obj)


def remove_role(user, role_class=None, obj=None):
    """
    Proxy method to be used for one
    User instance.
    """
    remove_roles([user], role_class, obj)


def remove_roles(users_list, role_class=None, obj=None):
    """
    Delete all RolePermission objects in the database
    referencing the followling role_class to the
    user.
    If "obj" is provided, only the instances refencing
    this object will be deleted.
    """
    query = UserRole.objects.filter(user__in=users_list)
    role = None

    if role_class:
        # Filtering by role class.
        role = get_roleclass(role_class)
        query = query.filter(role_class=role.get_class_name())

    if obj:
        # Filtering by object.
        ct_obj = ContentType.objects.get_for_model(obj)
        query = query.filter(content_type=ct_obj.id, object_id=obj.id)

    # Check if object belongs
    # to the role class.
    check_my_model(role, obj)

    # Cleaning the cache system.
    for user in users_list:
        delete_from_cache(user, obj)

    # Cleaning the database.
    query.delete()


def remove_all(role_class=None, obj=None):
    """
    Remove all roles of the project.

    If "role_class" is provided,
    only the roles of "role_class"
    will be affected.

    If "obj" is provided, only
    users for that object will
    lose the role.
    """
    query = UserRole.objects.all()
    role = None

    if role_class:
        # Filtering by role class.
        role = get_roleclass(role_class)
        query = UserRole.objects.filter(role_class=role.get_class_name())

    if obj:
        # Filtering by object.
        ct_obj = ContentType.objects.get_for_model(obj)
        query = query.filter(content_type=ct_obj.id, object_id=obj.id)

    # Check if object belongs
    # to the role class.
    check_my_model(role, obj)

    # Cleaning the cache system.
    for role_obj in query:
        delete_from_cache(role_obj.user, role_obj.obj)

        # Cleaning the database.
        role_obj.delete()



# Generated by Django 2.0.1 on 2018-02-27 21:08

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('improved_permissions', '0001_initial'),
        ('auth', '0009_alter_user_last_name_max_length'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.AddField(
            model_name='userrole',
            name='user',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='roles', to=settings.AUTH_USER_MODEL, verbose_name='Usuário'),
        ),
        migrations.AddField(
            model_name='rolepermission',
            name='permission',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='accesses', to='auth.Permission'),
        ),
        migrations.AddField(
            model_name='rolepermission',
            name='role',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='accesses', to='improved_permissions.UserRole'),
        ),
        migrations.AlterUniqueTogether(
            name='userrole',
            unique_together={('user', 'role_class', 'content_type', 'object_id')},
        ),
        migrations.AlterUniqueTogether(
            name='rolepermission',
            unique_together={('role', 'permission')},
        ),
    ]



""" permissions models """
from django.conf import settings
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.db import models

from improved_permissions.exceptions import RoleNotFound
from improved_permissions.roles import ALL_MODELS, ALLOW_MODE
from improved_permissions.utils import (get_permissions_list, get_roleclass,
                                        permission_to_string)


class UserRole(models.Model):
    """
    UserRole

    This model represents the relationship between
    a user instance of the project with any other
    Django model, according to the rules defined
    in the Role class.
    """
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        related_name='roles',
        verbose_name='Usuário'
    )

    permissions = models.ManyToManyField(
        Permission,
        through='RolePermission',
        related_name='roles',
        verbose_name='Permissões'
    )

    role_class = models.CharField(max_length=256)

    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE, null=True)
    object_id = models.PositiveIntegerField(null=True)
    obj = GenericForeignKey()

    class Meta:
        verbose_name = 'Role Instance'
        verbose_name_plural = 'Role Instances'
        unique_together = ('user', 'role_class', 'content_type', 'object_id')

    def __str__(self):
        role = get_roleclass(self.role_class)
        output = '{user} is {role}'.format(user=self.user, role=role.get_verbose_name())
        if self.obj:
            output += ' of {obj}'.format(obj=self.obj)
        return output

    @property
    def role(self):
        return get_roleclass(self.role_class)

    def get_verbose_name(self):
        return self.role.get_verbose_name()

    def clean(self):
        try:
            get_roleclass(self.role_class)
        except RoleNotFound:
            raise ValidationError(
                {'role_class': 'This string representation does not exist as a Role class.'}
            )

    def save(self, *args, **kwargs):  # pylint: disable=arguments-differ,unused-argument
        self.clean()
        super().save()

        # non-object roles does not have specific
        # permissions auto created.
        if self.role.models == ALL_MODELS:
            return

        all_perms = get_permissions_list(self.role.get_models())

        # Filtering the permissions based
        # on "allow" or "deny".
        role_instances = list()
        for perm in all_perms:
            access = None
            perm_s = permission_to_string(perm)
            if self.role.get_mode() == ALLOW_MODE:
                # [Allow Mode]
                # Put the access as "True" only for
                # the permissions in allow list.
                access = True if perm_s in self.role.allow else False
            else:
                # [Deny Mode]
                # Put the acces as "False" only for
                # the permissions in deny list.
                access = False if perm_s in self.role.deny else True
            role_instances.append(RolePermission(role=self, permission=perm, access=access))

        RolePermission.objects.bulk_create(role_instances)


class RolePermission(models.Model):
    """
    RolePermission

    This model has the function of performing
    the m2m relation between the Permission
    and the UserRole instances. It is possible
    that different instances of the same UserRole
    may have access to different permissions.
    """
    PERMISSION_CHOICES = (
        (True, 'Allow'),
        (False, 'Deny')
    )

    access = models.BooleanField(choices=PERMISSION_CHOICES, default=True)

    role = models.ForeignKey(
        UserRole,
        on_delete=models.CASCADE,
        related_name='accesses'
    )

    permission = models.ForeignKey(
        Permission,
        on_delete=models.CASCADE,
        related_name='accesses'
    )

    class Meta:
        unique_together = ('role', 'permission')



"""getters functions"""
from django.contrib.auth import get_user_model
from django.contrib.contenttypes.models import ContentType

from improved_permissions.exceptions import NotAllowed
from improved_permissions.models import UserRole
from improved_permissions.utils import check_my_model, get_roleclass


def get_user(role_class=None, obj=None):
    """
    Get the User instance attached to the object.
    Only one UserRole must exists and this relation
    must be unique=True.

    Returns None if there is no user attached
    to the object.
    """
    query = UserRole.objects.select_related('user').all()
    role = None

    if role_class:
        # All users who have "role_class" attached to any object.
        role = get_roleclass(role_class)
        query = query.filter(role_class=role.get_class_name())

    if obj:
        # All users who have any role attached to the object.
        ct_obj = ContentType.objects.get_for_model(obj)
        query = query.filter(content_type=ct_obj.id, object_id=obj.id)

    # Check if object belongs
    # to the role class.
    check_my_model(role, obj)

    # Looking for a role class using unique=True
    selected = list()
    for ur_obj in query:
        role = get_roleclass(ur_obj.role_class)
        if role.unique is True:
            selected.append(ur_obj.user)

    users_set = set(selected)
    if len(users_set) > 1:
        raise NotAllowed(
            'Multiple unique roles was found using '
            'the function get_user.  Use get_users '
            'instead.'
        )
    if len(users_set) == 1:
        return selected[0]
    return None


def get_users(role_class=None, obj=None):
    """
    If "role_class" and "obj" is provided,
    returns a QuerySet of users who has
    this role class attached to the
    object.

    If only "role_class" is provided, returns
    a QuerySet of users who has this role
    class attached to any object.

    If neither "role_class" or "obj" are provided,
    returns all users of the project.
    """

    query = get_user_model().objects.all()
    role = None

    if role_class:
        # All users who have "role_class" attached to any object.
        role = get_roleclass(role_class)
        query = query.filter(roles__role_class=role.get_class_name())

    if obj:
        # All users who have any role attached to the object.
        ct_obj = ContentType.objects.get_for_model(obj)
        query = query.filter(roles__content_type=ct_obj.id, roles__object_id=obj.id)

    # Check if object belongs
    # to the role class.
    check_my_model(role, obj)

    # Return as a distinct QuerySet.
    return query.distinct()


def get_objects(user, role_class=None, model=None):
    """
    Return the list of objects attached
    to a given user.

    If "role_class" is provided, only the objects
    which as registered in that role class will
    be returned.

    If "model" is provided, only the objects
    of that model will be returned.
    """
    query = UserRole.objects.filter(user=user)
    role = None

    if role_class:
        # Filtering by role class.
        role = get_roleclass(role_class)
        query = query.filter(role_class=role.get_class_name())

    if model:
        # Filtering by model.
        ct_obj = ContentType.objects.get_for_model(model)
        query = query.filter(content_type=ct_obj.id)

    # Check if object belongs
    # to the role class.
    check_my_model(role, model)

    # TODO
    result = set(ur_obj.obj for ur_obj in query)
    # TODO
    return list(result)


def get_role(user, obj=None):
    """
    Proxy method to be used when you sure that
    will have only one role class attached.
    """
    return get_roles(user, obj)[0]


def get_roles(user, obj=None):
    """
    Return a list of role classes
    that is attached to "user".

    If "obj" is provided, the object
    must be attached as well.
    """
    query = UserRole.objects.filter(user=user)
    if obj:
        ct_obj = ContentType.objects.get_for_model(obj)
        query = query.filter(content_type=ct_obj.id, object_id=obj.id)

    # Transform the string representations
    # into role classes and return as list.
    return [get_roleclass(ur_obj.role_class) for ur_obj in query]
